////////////////////////////////////////////////////////////////////////////
//  
//  KReptonEd - editor for Repton levels for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kmainwindow.h>
#include <kmessagebox.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kglobal.h>
#include <kconfig.h>
#include <klistbox.h>
#include <klineedit.h>
#include <knuminput.h>
#include <kpushbutton.h>

#include <qpainter.h>
#include <qfile.h>

#include "editorwidgetui.h"
#include "spriteeditor.h"
#include "mapeditor.h"
#include "newmapdialog.h"
#include "newtransporterdialog.h"
#include "checkmap.h"
#include "fileio.h"
#include "editorwindow.h"


EditorWindow::EditorWindow(const char *name)
	: KMainWindow(0,name)
{
	kdDebug(0) << "EditorWindow::EditorWindow name='" << name << "'" << endl;

	modified = true;				// force update first time
	caption = name;

	(void) KStdAction::close(this,SLOT(close()),actionCollection());
	(void) KStdAction::save(this,SLOT(menu_save()),actionCollection());

	(void) new KAction("Check Consistenc&y","check",Key_F10,
				    this,SLOT(menu_strictCheck()),
				    actionCollection(),"file_check");

	strictAction = new KToggleAction("Check Consistenc&y Before Saving",0,
					 actionCollection(),"settings_check");

	alignAction = new KToggleAction("&Align Windows",0,
					 actionCollection(),"settings_align");

	spriteToggle = new KToggleAction("&Sprite Editor",Key_F2,
					 this,SLOT(menu_spriteEditor()),
					 actionCollection(),"window_spriteeditor");

	mapToggle = new KToggleAction("&Level Editor",Key_F3,
				      this,SLOT(menu_levelEditor()),
				      actionCollection(),"window_leveleditor");

	realignAction = new KAction("&Realign Windows",Key_F4,
				    this,SLOT(menu_realign()),
				    actionCollection(),"window_realign");

	createGUI();
	toolBar("mainToolBar")->hide();

	view = new EditorWidget(this);
	setCentralWidget(view);
	view->nameLineEdit->setEnabled(false);

	connect(view->mapsListBox,SIGNAL(highlighted(int)),
		this,SLOT(selectedMap(int)));
	connect(view->passwordLineEdit,SIGNAL(textChanged(const QString&)),
		this,SLOT(changedPassword(const QString&)));
	connect(view->timeSpinBox,SIGNAL(valueChanged(int)),
		this,SLOT(changedTime(int)));
	connect(view->newlevelPushButton,SIGNAL(clicked()),
		this,SLOT(mapInsert()));
	connect(view->removelevelPushButton,SIGNAL(clicked()),
		this,SLOT(mapRemove()));
	connect(view->levelupPushButton,SIGNAL(clicked()),
		this,SLOT(mapMoveUp()));
	connect(view->leveldownPushButton,SIGNAL(clicked()),
		this,SLOT(mapMoveDown()));
	connect(view->newtransportPushButton,SIGNAL(clicked()),
		this,SLOT(transporterInsert()));
	connect(view->changetransportPushButton,SIGNAL(clicked()),
		this,SLOT(transporterChange()));
	connect(view->removetransportPushButton,SIGNAL(clicked()),
		this,SLOT(transporterRemove()));
	connect(view->checkPushButton,SIGNAL(clicked()),
		this,SLOT(menu_strictCheck()));

	sprites = new QPixmap[num_sprites];
	sprite_window = new SpriteEditor(sprites,this);
	connect(sprite_window,SIGNAL(changedSprite()),
		this,SLOT(changedSprite()));
	connect(sprite_window,SIGNAL(closed()),
		this,SLOT(updateWindowStates()));

	map_window = new MapEditor(sprites,this);
	connect(map_window,SIGNAL(modified(bool)),
		this,SLOT(setModified(bool)));
	connect(map_window,SIGNAL(closed()),
		this,SLOT(updateWindowStates()));

	readOptions();
	setModified(false);
	updateWindowStates();

	finalizeGUI(false);
	setFixedSize(sizeForCentralWidgetSize(view->size()));
	kdDebug(0) << "EditorWindow::EditorWindow done" << endl;
}


bool EditorWindow::loadEpisode(const QString& path)
{
	Map *map;
	maps.clear();

	QString filename = path + "/info";
	QString name;
	int num_levels;

	QFile f(filename);
	if (!f.open(IO_ReadOnly)) return (false);

	QTextStream t(&f);
	name = t.readLine();
	t >> num_levels;
	f.close();

	view->nameLineEdit->setText(name);

	for (int i = 0; i<num_levels; ++i)
	{
		filename = QString("%1/map%2").arg(path).arg(i+1);
		map = new Map();
		if (!map->load(filename)) return (false);
		maps.append(map);
	}

	updateMapsList();
	view->mapsListBox->setCurrentItem(0);

	if (!loadSprites(path)) return (false);

	setModified(false);
	return (true);
}


bool EditorWindow::loadSprites(const QString& path)
{
	QPixmap allsprites;

	QString filename = path+"/sprites.bmp";
	if (!allsprites.load(filename))
	{
		FileIO::report("Cannot load sprites from '%1'",filename);
		return (false);
	}

	for (int y = 0; y<source_height; ++y)
	{
		for (int x = 0; x<source_width; ++x)
		{
			int i = y*source_width + x; 
			sprites[i].resize(sprite_width,sprite_height);

			QPainter p;
			p.begin(&sprites[i]);
			p.drawPixmap(0,0,allsprites,
				     x*sprite_width,y*sprite_height,
				     sprite_width,sprite_height);
			p.end();
		}
	}

	return (true);
}


void EditorWindow::updateMapsList()
{
	view->mapsListBox->setAutoUpdate(false);
	view->mapsListBox->clear();
	for (Map *m = maps.first(); m != 0; m = maps.next())
		view->mapsListBox->insertItem(m->password);
	view->mapsListBox->setAutoUpdate(true);

	view->removelevelPushButton->setEnabled(maps.count()>1);
}


void EditorWindow::updateTransportersList(int level)
{
	Map *map = maps.at(level);

	view->transportListBox->setAutoUpdate(false);
	view->transportListBox->clear();
	for (Transporter *t = map->transporters.first(); t != 0;
	     t = map->transporters.next()) {
		QString s;
		s.sprintf("%d,%d -> %d,%d",
			  t->orig_x+1, t->orig_y+1, t->dest_x+1, t->dest_y+1);
		view->transportListBox->insertItem(s);
	}

	const bool any = (map->transporters.count()>0);
	if (any)
	{
		view->transportListBox->setCurrentItem(0);
		view->transportListBox->centerCurrentItem();
	}
	view->removetransportPushButton->setEnabled(any);
	view->changetransportPushButton->setEnabled(any);

	view->transportListBox->setAutoUpdate(true);
}


void EditorWindow::selectLevel(int level)
{
	if (level<0)
	{
		view->passwordLineEdit->setEnabled(false);
		view->timeSpinBox->setEnabled(false);
		return;
	}

	Map *map = maps.at(level);
	map_window->setMap(map);

	updateTransportersList(level);
	view->passwordLineEdit->blockSignals(true);
	view->passwordLineEdit->setText(map->password);
	view->passwordLineEdit->blockSignals(false);
	view->timeSpinBox->setValue(map->time_limit);

	++level;
	view->levelupPushButton->setEnabled(level>1);
	view->leveldownPushButton->setEnabled(level<((int) maps.count()));
	view->passwordLineEdit->setEnabled(true);
	view->timeSpinBox->setEnabled(true);
}


void EditorWindow::alignWindows()
{
	QRect g = frameGeometry();
	sprite_window->move(x() + g.width(), y());
	QRect h = sprite_window->frameGeometry();
	map_window->move(x() + g.width(), sprite_window->y() + h.height());
}


void EditorWindow::mapInsert()
{
	NewMapDialog d(this);

	int sx,sy;
	for (;;)
	{
		if (!d.exec()) return;
// todo: can validate non-blank on the fly?
		if (d.mapPassword()=="")
		{
			KMessageBox::sorry(this,"You must specify a name/password for the level");
			continue;
		}

		sx = d.mapWidth();
		sy = d.mapHeight();
		if (sx<10 || sy<10)			// shouldn't happen, spinboxes enforce
		{
			KMessageBox::error(this,"Invalid map size");
			continue;
		}
		break;
	}

	Map *m = new Map();
	m->newMap(sx,sy);
	m->password = d.mapPassword();
	maps.append(m);

	updateMapsList();
	view->mapsListBox->setCurrentItem(maps.count()-1);
	view->mapsListBox->centerCurrentItem();
	setModified();
}


void EditorWindow::mapRemove()
{
	if (maps.count()==1)				// shouldn't happen, GUI disables
	{
		KMessageBox::error(this,"Cannot remove the only level");
		return;
	}

	int item = view->mapsListBox->currentItem();
	QString msg;
	msg.sprintf("Are you sure to remove the map #%d?",item+1);

	if (KMessageBox::warningContinueCancel(this,msg,QString::null,
					       KGuiItem("Remove"))==KMessageBox::Continue)
	{
		Map *m = maps.at(item);
		maps.remove(item);
		delete m;

		updateMapsList();
		view->mapsListBox->setCurrentItem(0);
		view->mapsListBox->centerCurrentItem();
		setModified();
	}
}


void EditorWindow::mapMoveUp()
{
	int item = view->mapsListBox->currentItem();
	if (item<=0)					// shouldn't happen, GUI disables
	{
		KMessageBox::error(this,"Cannot move up the first level");
		return;
	}

	Map *m = maps.at(item);
	maps.remove(item);
	maps.insert(item-1,m);

	updateMapsList();
	view->mapsListBox->setCurrentItem(item-1);
	view->mapsListBox->centerCurrentItem();
	setModified();
}


void EditorWindow::mapMoveDown()
{
	int item = view->mapsListBox->currentItem();
	int last_item = maps.count()-1;
	if (item>=last_item)				// shouldn't happen, GUI disables
	{
		KMessageBox::error(this,"Cannot move down the last level");
		return;
	}

	Map *m = maps.at(item);
	maps.remove(item);
	maps.insert(item+1,m);

	updateMapsList();
	view->mapsListBox->setCurrentItem(item+1);
	view->mapsListBox->centerCurrentItem();
	setModified();
}


void EditorWindow::transporterInsert()
{
	Map *map = maps.at(view->mapsListBox->currentItem());

	NewTransporterDialog d("New Transporter",this);
	d.setLimits(map->width,map->height);
	d.setValues(1,1,2,2);

	int ox,oy,dx,dy;
	for (;;)
	{
		if (!d.exec()) return;

		ox = d.xorig();
		oy = d.yorig();
		dx = d.xdest();
		dy = d.ydest();

		if (ox<1 || oy<1 || dx<1 || dy<1 ||	// shouldn't happen, spinboxes enforce
		    ox>map->width || oy>map->height ||
		    dx>map->width || dy>map->height)
		{
			KMessageBox::error(this,"Invalid transporter coordinates");
			continue;
		}
		break;
	}

	Transporter *t = new Transporter(ox-1,oy-1,dx-1,dy-1);
	map->transporters.append(t);

	updateTransportersList(view->mapsListBox->currentItem());
	view->transportListBox->setCurrentItem(map->transporters.count() - 1);
	view->transportListBox->centerCurrentItem();
	setModified();
}


void EditorWindow::transporterChange()
{
	int item = view->transportListBox->currentItem();
	Map *map = maps.at(view->mapsListBox->currentItem());

	Transporter *t = map->transporters.at(item);
	int ox,oy,dx,dy;
	ox = t->orig_x+1;
	oy = t->orig_y+1;
	dx = t->dest_x+1;
	dy = t->dest_y+1;

	NewTransporterDialog d("Change Transporter",this);
	d.setLimits(map->width,map->height);
	d.setValues(ox,oy,dx,dy);

	for (;;)
	{
		if (!d.exec()) return;

		ox = d.xorig();
		oy = d.yorig();
		dx = d.xdest();
		dy = d.ydest();

		if (ox<1 || oy<1 || dx<1 || dy<1 ||	// shouldn't happen, spinboxes enforce
		    ox>map->width || oy>map->height ||
		    dx>map->width || dy>map->height)
		{
			KMessageBox::error(this,"Invalid transporter coordinates");
			continue;
		}
		break;
	}

	t->orig_x = ox-1;
	t->orig_y = oy-1;
	t->dest_x = dx-1;
	t->dest_y = dy-1;

	updateTransportersList(view->mapsListBox->currentItem());
	view->transportListBox->setCurrentItem(item);
	view->transportListBox->centerCurrentItem();
	setModified();
}


void EditorWindow::transporterRemove()
{
	Map *map = maps.at(view->mapsListBox->currentItem());

	if (map->transporters.count()<1)		// shouldn't happen, GUI disables
	{
		KMessageBox::error(this,"No transporters to remove");
		return;
	}

	int item = view->transportListBox->currentItem();
	QString msg;
	msg.sprintf("Are you sure to remove the transporter #%d?",item+1);

	if (KMessageBox::warningContinueCancel(this,msg,QString::null,
					       KGuiItem("Remove"))==KMessageBox::Continue)
	{
		Transporter *t = map->transporters.at(item);
		map->transporters.remove(item);
		delete t;
		updateTransportersList(view->mapsListBox->currentItem());
		setModified();
	}
}


void EditorWindow::selectedMap(int level)
{
	selectLevel(level);
}


void EditorWindow::changedPassword(const QString &s)
{
	int level = view->mapsListBox->currentItem();
	if (level >= 0)
	{
		maps.at(level)->password = s;
		view->mapsListBox->changeItem(s,level);
	}
	setModified();
}


void EditorWindow::changedTime(int i)
{
	int level = view->mapsListBox->currentItem();
	if (level>=0) maps.at(level)->time_limit = i;
	setModified();
}


void EditorWindow::changedSprite()
{
	sprite_window->updateChilds();
	map_window->updateChilds();
	setModified();
}


void EditorWindow::menu_save()
{
	const QString name = view->nameLineEdit->text();
	if (name=="")					// shouldn't happen, not editable
	{
		KMessageBox::sorry(this,"Unable to save this episode; it has no name");
		return;
	}

	if (strictAction->isChecked())
	{
		if (CheckMap::checkReport(maps,this,
					  "Unable to save this episode",
					  false)==CheckMap::Fatal) return;
	}

	if (FileIO::saveEpisode(name,maps,sprites)) setModified(false);
}


void EditorWindow::menu_strictCheck()
{
	CheckMap::checkReport(maps,this,QString::null,true);
}


void EditorWindow::menu_spriteEditor()
{
	if (!spriteToggle->isChecked()) sprite_window->hide();
	else
	{
		sprite_window->show();
		if (alignAction->isChecked()) alignWindows();
	}

	updateWindowStates();
}


void EditorWindow::menu_levelEditor()
{
	if (!mapToggle->isChecked()) map_window->hide();
	else
	{
		map_window->show();
		if (alignAction->isChecked()) alignWindows();
	}

	updateWindowStates();
}


void EditorWindow::menu_realign()
{
	alignWindows();
}


void EditorWindow::moveEvent(QMoveEvent *e)
{
	KMainWindow::moveEvent(e);
	if (alignAction->isChecked()) alignWindows();
}


void EditorWindow::setModified(bool mod)
{
	if (modified==mod) return;			// no change
	modified = mod;

	const QString name = view->nameLineEdit->text();
	setCaption(QString("%1 '%2'").arg(caption).arg(name),modified);
}


bool EditorWindow::queryClose()
{
	if (!modified) return (true);			// can close now

	const QString name = view->nameLineEdit->text();
	switch (KMessageBox::warningYesNoCancel(this,
						QString("Save changes to episode '%1'?").arg(name)))
	{
case KMessageBox::Yes:
		menu_save();				// save the episode
		return (!modified);			// check it succeeded

case KMessageBox::No:
		return (true);				// discard

default:	return (false);				// cancel
	}
}


void EditorWindow::updateWindowStates()
{
	const bool svis = sprite_window->isVisible();
	const bool mvis = map_window->isVisible();

	spriteToggle->setChecked(svis);
	mapToggle->setChecked(mvis);
	realignAction->setEnabled(svis || mvis);
}


void EditorWindow::readOptions()
{
	kdDebug(0) << "EditorWindow::readOptions()" << endl;

	KConfig *config = KGlobal::config();
	config->setGroup("Options");

	strictAction->setChecked(config->readBoolEntry("StrictChecking",true));
	alignAction->setChecked(config->readBoolEntry("AutoAlign",true));
}


void EditorWindow::saveOptions()
{
	kdDebug(0) << "EditorWindow::saveOptions()" << endl;

	KConfig *config = KGlobal::config();
	config->setGroup("Options");

	config->writeEntry("StrictChecking",strictAction->isChecked());
	config->writeEntry("AutoAlign",alignAction->isChecked());
}


void EditorWindow::closeEvent(QCloseEvent *e)
{
	saveOptions();
	KMainWindow::closeEvent(e);
}
