////////////////////////////////////////////////////////////////////////////
//  
//  KReptonEd - editor for Repton levels for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kapplication.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <kmessagebox.h>

#include <qdir.h>
#include <qfile.h>
#include <qtstream.h>
#include <qlist.h>
#include <qpainter.h>

#include <errno.h>
#include <string.h>

#include "kreptoned.h"
#include "editormap.h"
#include "fileio.h"


static void removeMapFiles(const QString& path)
{
	QDir dir(path);
	if (!dir.exists())
	{
		FileIO::report("Cannot read directory '%1'",path);
		return;
	}

	dir.setFilter(QDir::Files);
	dir.setNameFilter("map*");
	dir.setSorting(QDir::Name);
	const QFileInfoList *list = dir.entryInfoList();
	QFileInfoListIterator it(*list);
	QFileInfo *fi;
	while ((fi = it.current())!=0)
	{
		dir.remove(fi->fileName());
		++it;
	}
}


static bool createPath(const QString& path)
{
	// Create the destination directory
	QString filename;
	QDir dir(path);
	if (!dir.exists()) KStandardDirs::makeDir(path);

	dir.cd(path);
	if (!dir.exists())
	{
		FileIO::report("Cannot create directory '%1'",path);
		return (false);
	}
	return (true);
}


// todo: use kioslave?
static bool copyFile(const QString& srcfile,const QString& destfile)
{
	QFile fin(srcfile);
	if (!fin.open(IO_ReadOnly))
	{
		FileIO::report("Cannot read from file '%1'",srcfile);
		return (false);
	}

	QFile fout(destfile);
	if (!fout.open(IO_WriteOnly))
	{
		FileIO::report("Cannot write to file '%1'",destfile);
		return (false);
	}

	char buf[1024];
	int i;
	while ((i = fin.readBlock(buf,1024))>0) fout.writeBlock(buf,i);

	fin.close();
	fout.close();
	return (true);
}


static QString makeFilename(const QString &s)
{
	QString s1 = s.simplifyWhiteSpace();
	QString t = "";

	QChar ch;
	for (int i = 0; !((ch = s1.at(i)).isNull()); ++i)
	{
		if (ch.latin1()==0 || !ch.isLetterOrNumber()) ch = '_';
		t += ch.lower();
	}

//	kdDebug(0) << "FileIO::makeFilename(" << s << ") -> '" << t << "'" << endl;
	return (t);
}


void FileIO::loadEpisodeList(QStringList& l,FileIO::EpisodeType what)
{
	QString path;
	bool loaded = false;
	const QString localdir = QDir(KGlobal::dirs()->localkdedir()).canonicalPath()+"/";

	QStringList dirs = KGlobal::dirs()->findDirs("episodes","");

	kdDebug(0) << "loadEpisodeList: local='" << localdir << "' what=" << what << endl;
	for (QStringList::Iterator di = dirs.begin(); di!=dirs.end(); ++di)
	{
		QString di1 = *di;

		kdDebug(0) << "loadEpisodeList: dir='" << di1 << "'" << endl;
		if (di1==QString::null) continue;
		di1 = QDir(di1).canonicalPath();
		if (di1.startsWith(localdir))
		{
			if (what==FileIO::Global) continue;
		}
		else
		{
			if (what==FileIO::Local) continue;
		}

		QDir dir(di1);
		if (!dir.exists())
		{
			FileIO::report("Cannot read episode directory '%1'",path);
			continue;
		}
 
		dir.setFilter(QDir::Dirs);
		dir.setSorting(QDir::Name);
		const QFileInfoList *list = dir.entryInfoList();
		QFileInfoListIterator it(*list);
		QFileInfo *fi;

		while ((fi = it.current())!=0)
		{
			if (!fi->fileName().startsWith("."))
			{
				if (!dir.cd(fi->fileName()))
				{
					FileIO::report("Cannot access directory '%1'",
					      dir.absFilePath(fi->fileName()));
					continue;
				}

				QString filename = dir.absPath()+"/info";
				QFile f(filename);
				if (!f.open(IO_ReadOnly))
				{
					FileIO::report("Cannot read information file '%1'",
						       filename);
					continue;
				}

				QTextStream t(&f);
				QString name;
				name = t.readLine().stripWhiteSpace();
				f.close();
				dir.cdUp();

				if (name=="blank") l.prepend(name);
				else l.append(name);	// force "blank" to front
				loaded = true;
			}
			++it;
		}
	}

	if (what!=FileIO::Local && l.count()==0) FileIO::report("Cannot find any episodes.<br>Is the application properly installed?",
								QString::null,false,true);
}


int FileIO::countUserEpisodes()
{
	QStringList l;
	loadEpisodeList(l,FileIO::Local);
	return l.count();
}


void FileIO::removeEpisode(const QString& path)
{
	QDir dir(path);
	if (!dir.exists())
	{
		FileIO::report("Cannot access directory '%1'",path);
		return;
	}

	QString dirname = dir.dirName();
	dir.setFilter(QDir::Files);
	dir.setSorting(QDir::Name);
	const QFileInfoList *list = dir.entryInfoList();
	QFileInfoListIterator it(*list);
	QFileInfo *fi;
	while ((fi = it.current()) != 0)
	{
		if (!dir.remove(fi->fileName()))
		{
			FileIO::report("Cannot delete file '%1'",
				       dir.absFilePath(fi->fileName()));
			break;
		}
		++it;
	}

	dir.cdUp();
	if (!dir.rmdir(path)) FileIO::report("Cannot delete directory '%1'",path);
}


bool FileIO::saveEpisode(const QString &name,
			 QList<Map> maps,const QPixmap *sprites)
{
	QString path = KGlobal::dirs()->saveLocation("episodes");

	kdDebug(0) << "FileIO::saveEpisode name='" << name << "' path='" << path << "'" << endl;
	if (path==QString::null) return (false);

	path += makeFilename(name);
	kdDebug(0) << "FileIO::saveEpisode full path='" << path << "'" << endl;

	QString filename;

	filename = path+"/info";
	QFile f(filename);
	if (!f.open(IO_WriteOnly))
	{
		FileIO::report("Cannot write to file '%1'",filename);
		return (false);
	}

	QTextStream t(&f);
	t << name << '\n';
	t << maps.count() << '\n';
	f.close();

	removeMapFiles(path);

	for (Map *map = maps.first(); map != 0; map = maps.next())
	{
		filename = QString("%1/map%2").arg(path).arg(maps.at()+1);
		if (!map->save(filename))
		{
			FileIO::report("Cannot save map to '%1'",filename);
			return (false);
		}
	}

	QPixmap allsprites(source_width*sprite_width,source_height*sprite_height);

	QPainter p;
	p.begin(&allsprites);
	for (int y = 0; y<source_height; ++y)
	{
		for (int x = 0; x<source_width; ++x)
		{
//			p.drawPixmap(x*source_width,y*source_height,
			p.drawPixmap(x*sprite_width,y*sprite_height,
				     sprites[y*source_width+x]);
		}
	}
	p.end();

	if (!allsprites.save(path+"/sprites.bmp","BMP"))
	{
		FileIO::report("Cannot save sprites to '%1'",filename);
		return (false);
	}

	return (true);
}


// todo: do we even need a loop here?
QString FileIO::locateEpisode(const QString &name,FileIO::EpisodeType what)
{
	kdDebug(0) << "FileIO::locateEpisode name='" << name << "'" << endl;

	QStringList l;
	loadEpisodeList(l,what);

	QString path = QString::null;
	for (QStringList::iterator it = l.begin(); it!=l.end(); ++it)
	{
		if (name==(*it))
		{
			path = FileIO::locateEpisodeDir(name);
			break;
		}
	}

	kdDebug(0) << "FileIO::locateEpisode path='" << path << "'" << endl;
	return (path);
}


QString FileIO::locateEpisodeDir(const QString &name)
{
	const QString filename = makeFilename(name);
	QString path;

	path = KGlobal::dirs()->findResource("episodes",filename+"/info");
	if (path!=QString::null) path = path.left(path.length()-5);
	return (path);
}


QString FileIO::copyEpisode(const QString &name,const QString &srcpath)
{
	QString destpath;

	if (srcpath==QString::null) return (QString::null);

	QString filename = makeFilename(name);
	destpath = KGlobal::dirs()->saveLocation("episodes",filename+"/");
	kdDebug(0) << "FileIO::copyEpisode name='" << name << "' srcpath='" << srcpath << "' destpath='" << destpath << "'" << endl;
	if (!createPath(destpath)) return (QString::null);

	int levels;

	filename = srcpath+"/info";
	QFile fin(filename);
	if (!fin.open(IO_ReadOnly))
	{
		FileIO::report("Cannot read from source file '%1'",filename);
		return (QString::null);
	}

	QTextStream tin(&fin);
	(void) tin.readLine();				// drop old name
	tin >> levels;					// read count of levels
	fin.close();

	filename = destpath+"/info";
	QFile fout(filename);
	if (!fout.open(IO_WriteOnly))
	{
		FileIO::report("Cannot write to destination file '%1'",filename);
		return (QString::null);
	}

	QTextStream tout(&fout);
	tout << name << '\n';				// save new name
	tout << levels << '\n';				// copy count of levels
	fout.close();

	if (!copyFile(srcpath+"/sprites.bmp",destpath+"/sprites.bmp")) return (QString::null);
	for (int i = 0; i<levels; ++i)
	{
		filename.sprintf("/map%d",i+1);
		if (!copyFile(srcpath+filename,destpath+filename)) return (QString::null);
	}

	return (destpath);
}


void FileIO::report(const QString &message,const QString &filename,
		    bool system,bool fatal)
{
	QString s = QString("<qt>%1").arg(message);
	if (filename!=QString::null) s = s.arg(filename);
	if (system)
	{
		s += "<p>The error reported was: ";
		s += strerror(errno);
	}

	if (fatal)
	{
		KMessageBox::error(0,s);
		kapp->exit(1);
	}
	else KMessageBox::sorry(0,s);
}
