////////////////////////////////////////////////////////////////////////////
//  
//  KReptonEd - editor for Repton levels for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kapplication.h>

#include <qlabel.h>
#include <qlistbox.h>
#include <qpainter.h>
#include <qpushbt.h>
#include <qscrbar.h>
#include <qtooltip.h>

#include "kreptoned.h"
#include "editormap.h"
#include "mapeditor.h"
#include "spritepreviewframe.h"
#include "mapgridframe.h"


const int default_width = 555;
const int default_height = 355;


MapEditor::MapEditor(QPixmap *xsprites, QWidget *parent, const char *name)
	: QWidget(parent, name,Qt::WType_TopLevel)
{
	const char *sprite_names[] =
	{
#define SPRITEOBJ(e, n)
#define MAPOBJ(e, n) n,
#include "sprites.h"
#undef SPRITEOBJ
#undef MAPOBJ
		0
	};

	setCaption("Level Editor");
	setMinimumSize(default_width, default_height);
	resize(default_width, default_height);
	setSizeIncrement(32, 32);

	map = 0;
	sprites = xsprites;
	current_sprite = Sprite_Empty;

	sprite_list = new QListBox(this);
	sprite_list->setGeometry(10, 30, 150, 260);
	sprite_list->setAutoUpdate(false);
	for (const char **s = &sprite_names[0]; *s != 0; ++s)
		sprite_list->insertItem(*s);
	sprite_list->setAutoUpdate(true);
	sprite_list->repaint();
	QToolTip::add(sprite_list, "The list of available objects to insert");

	QLabel *label = new QLabel(sprite_list, "&Object list", this);
	label->setAutoResize(true);
	label->move(10, 10);

	sprite_frame = new SpritePreviewFrame(sprites, this);
	sprite_frame->move(172, 30);
	QToolTip::add(sprite_frame, "The current selected object");
	label = new QLabel(sprite_frame, "Object", this);
	label->setAutoResize(true);
	label->move(172, 10);

	map_frame = new MapGridFrame(sprites, this);
	map_frame->move(225, 30);
	label = new QLabel(map_frame, "Level", this);
	label->setAutoResize(true);
	label->move(225, 10);

	h_sbar = new QScrollBar(QScrollBar::Horizontal, this);
	h_sbar->setGeometry(map_frame->x(),
			    map_frame->y() + map_frame->height(),
			    map_frame->width(), 16);
	v_sbar = new QScrollBar(QScrollBar::Vertical, this);
	v_sbar->setGeometry(map_frame->x() + map_frame->width(),
			    map_frame->y(),
			    16, map_frame->height());

	close_button = new QPushButton("&Close", this);
	close_button->setGeometry(10, 310, 70, 30);
	close_button->setDefault(true);
	QToolTip::add(close_button, "Close this window");

	coord = new QLabel("X:1 Y:1", this);
	coord->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	coord->setAutoResize(true);
	coord->move(map_frame->x() + map_frame->width() - coord->width(), 3);

        sprite_list->setFocus();

	connect(close_button, SIGNAL(clicked()),
		this, SLOT(hide()));
	connect(sprite_list, SIGNAL(highlighted(int)),
		this, SLOT(selectedSprite(int)));

	connect(map_frame, SIGNAL(pressedButton(int, int, int)),
		this, SLOT(pressedButton(int, int, int)));
	connect(map_frame, SIGNAL(changedCoordinates(int, int)),
		this, SLOT(updateCoordinates(int, int)));
	connect(h_sbar, SIGNAL(valueChanged(int)),
		this, SLOT(hsbarChanged(int)));
	connect(v_sbar, SIGNAL(valueChanged(int)),
		this, SLOT(vsbarChanged(int)));
}


void MapEditor::setMap(Map *xmap)
{
	map = xmap;

	resize(default_width, default_height);
	map_frame->setSize(9,9);

	setMaximumSize(default_width+(map->width-map_frame->map_width)*32,
		       default_height+(map->height-map_frame->map_height)*32);
	h_sbar->setValue(0);
	h_sbar->setRange(0, map->width - map_frame->map_width);
	h_sbar->setSteps(1, map_frame->map_width);
	v_sbar->setValue(0);
	v_sbar->setRange(0, map->height - map_frame->map_height);
	v_sbar->setSteps(1, map_frame->map_height);
	map_frame->setMap(map);
	map_frame->repaint(false);
}


void MapEditor::hsbarChanged(int i)
{
	map_frame->map_sx = i;
	map_frame->repaint(false);
}


void MapEditor::vsbarChanged(int i)
{
	map_frame->map_sy = i;
	map_frame->repaint(false);
}


void MapEditor::selectedSprite(int i)
{
	current_sprite = (Sprite)i;
	sprite_frame->setSprite(current_sprite);
	sprite_frame->repaint(false);
}


void MapEditor::pressedButton(int button, int x, int y)
{
	if (button & LeftButton) map->xy(x, y) =  current_sprite;
	else if (button & RightButton) map->xy(x, y) =  Sprite_Empty;
	else return;

	map_frame->repaint(false);
	emit modified(true);
}


void MapEditor::updateCoordinates(int x, int y)
{
	QString s;
	s.sprintf("X:%d Y:%d", x+1, y+1);
	coord->setText(s);
	coord->move(map_frame->x() + map_frame->width() - coord->width(), 3);
}


void MapEditor::updateChilds()
{
	map_frame->repaint(false);
	sprite_frame->repaint(false);
}


void MapEditor::resizeEvent(QResizeEvent *)
{
	if (map==0) return;

	int xd, yd;
	xd = width() - default_width;
	yd = height() - default_height;
	map_frame->setSize(9 + xd/32, 9 + yd/32);
	h_sbar->setGeometry(map_frame->x(),
			    map_frame->y() + map_frame->height(),
			    map_frame->width(), 16);
	v_sbar->setGeometry(map_frame->x() + map_frame->width(),
			    map_frame->y(),
			    16, map_frame->height());
	h_sbar->setValue(0);
	map_frame->map_sx = 0;

	h_sbar->setRange(0, map->width - map_frame->map_width);
	h_sbar->setSteps(1, map_frame->map_width);
	v_sbar->setValue(0);
	map_frame->map_sy = 0;
	v_sbar->setRange(0, map->height - map_frame->map_height);
	v_sbar->setSteps(1, map_frame->map_height);

	close_button->move(close_button->x(), 310 + yd);
	coord->move(map_frame->x() + map_frame->width() - coord->width(), 3);
	sprite_list->resize(sprite_list->width(), 260 + yd);
	map_frame->repaint(false);
}


void MapEditor::closeEvent(QCloseEvent *e)
{
	QWidget::closeEvent(e);
	hide();
	emit closed();
}
